import csv
import itertools
import sys
import numpy

PROBS = {

    # Unconditional probabilities for having gene
    "gene": {
        2: 0.01,
        1: 0.03,
        0: 0.96
    },

    "trait": {

        # Probability of trait given two copies of gene
        2: {
            True: 0.65,
            False: 0.35
        },

        # Probability of trait given one copy of gene
        1: {
            True: 0.56,
            False: 0.44
        },

        # Probability of trait given no gene
        0: {
            True: 0.01,
            False: 0.99
        }
    },

    # Mutation probability
    "mutation": 0.01
}


def main():

    # Check for proper usage
    if len(sys.argv) != 2:
        sys.exit("Usage: python heredity.py data.csv")
    people = load_data(sys.argv[1])

    # Keep track of gene and trait probabilities for each person
    probabilities = {
        person: {
            "gene": {
                2: 0,
                1: 0,
                0: 0
            },
            "trait": {
                True: 0,
                False: 0
            }
        }
        for person in people
    }

    # Loop over all sets of people who might have the trait
    names = set(people)
    for have_trait in powerset(names):

        # Check if current set of people violates known information
        fails_evidence = any(
            (people[person]["trait"] is not None and
             people[person]["trait"] != (person in have_trait))
            for person in names
        )
        if fails_evidence:
            continue

        # Loop over all sets of people who might have the gene
        for one_gene in powerset(names):
            for two_genes in powerset(names - one_gene):

                # Update probabilities with new joint probability
                p = joint_probability(people, one_gene, two_genes, have_trait)
                update(probabilities, one_gene, two_genes, have_trait, p)

    # Ensure probabilities sum to 1
    normalize(probabilities)

    # Print results
    for person in people:
        print(f"{person}:")
        for field in probabilities[person]:
            print(f"  {field.capitalize()}:")
            for value in probabilities[person][field]:
                p = probabilities[person][field][value]
                print(f"    {value}: {p:.4f}")


def load_data(filename):
    """
    Load gene and trait data from a file into a dictionary.
    File assumed to be a CSV containing fields name, mother, father, trait.
    mother, father must both be blank, or both be valid names in the CSV.
    trait should be 0 or 1 if trait is known, blank otherwise.
    """
    data = dict()
    with open(filename) as f:
        reader = csv.DictReader(f)
        for row in reader:
            name = row["name"]
            data[name] = {
                "name": name,
                "mother": row["mother"] or None,
                "father": row["father"] or None,
                "trait": (True if row["trait"] == "1" else
                          False if row["trait"] == "0" else None)
            }
    return data


def powerset(s):
    """
    Return a list of all possible subsets of set s.
    """
    s = list(s)
    return [
        set(s) for s in itertools.chain.from_iterable(
            itertools.combinations(s, r) for r in range(len(s) + 1)
        )
    ]


def joint_probability(people, one_gene, two_genes, have_trait):
    """
    Compute and return a joint probability.

    The probability returned should be the probability that
        * everyone in set `one_gene` has one copy of the gene, and
        * everyone in set `two_genes` has two copies of the gene, and
        * everyone not in `one_gene` or `two_gene` does not have the gene, and
        * everyone in set `have_trait` has the trait, and
        * everyone not in set` have_trait` does not have the trait.
    """

    joint_probs = []

    # define likelihoods for better understanding
    only_through_mutation = PROBS['mutation']
    fifty_fifty = 0.5
    only_not_through_mutation = (1 - PROBS['mutation'])

    for person in people:
        # Get parents
        parents = [people[person]['mother'], people[person]['father']]

        # ------------------------  person has no gene  ----------------------------
        if person not in one_gene and person not in two_genes:
            # no information on parents
            if parents[0] is None:
                likelihood_g = PROBS['gene'][0]

            # with information on parents, iterate through options
            else:
                if parents[0] in two_genes:
                    if parents[1] in two_genes:
                        likelihood_g = (only_through_mutation * only_through_mutation)
                    elif parents[1] in one_gene:
                        likelihood_g = (only_through_mutation * fifty_fifty)
                    else:
                        likelihood_g = (only_through_mutation * only_not_through_mutation)
                elif parents[0] in one_gene:
                    if parents[1] in two_genes:
                        likelihood_g = (fifty_fifty * only_through_mutation)
                    elif parents[1] in one_gene:
                        likelihood_g = (fifty_fifty * fifty_fifty)
                    else:
                        likelihood_g = (fifty_fifty * only_not_through_mutation)
                else:
                    if parents[1] in two_genes:
                        likelihood_g = (only_not_through_mutation * only_through_mutation)
                    elif parents[1] in one_gene:
                        likelihood_g = (only_not_through_mutation * fifty_fifty)
                    else:
                        likelihood_g = (only_not_through_mutation * only_not_through_mutation)

            # trait likelihood for 0 genes
            if person not in have_trait:
                likelihood_t = PROBS['trait'][0][False]
            else:
                likelihood_t = PROBS['trait'][0][True]

        # ------------------------ person has one gene  ----------------------------
        elif person in one_gene:
            # no information on parents
            if parents[0] is None:
                likelihood_g = (PROBS['gene'][1])

            # with information on parents, iterate over options
            else:
                # ------------------------  from mom  ----------------------------
                if parents[0] in one_gene:
                    if parents[1] in one_gene:
                        from_mom = (fifty_fifty * fifty_fifty)
                    elif parents[1] in two_genes:
                        from_mom = (fifty_fifty * only_through_mutation)
                    else:
                        from_mom = (fifty_fifty * only_not_through_mutation)
                elif parents[0] in two_genes:
                    if parents[1] in one_gene:
                        from_mom = (only_not_through_mutation * fifty_fifty)
                    elif parents[1] in two_genes:
                        from_mom = (only_not_through_mutation * only_through_mutation)
                    else:
                        from_mom = (only_not_through_mutation * only_not_through_mutation)
                else:
                    if parents[1] in one_gene:
                        from_mom = (only_through_mutation * fifty_fifty)
                    elif parents[1] in two_genes:
                        from_mom = (only_through_mutation * only_through_mutation)
                    else:
                        from_mom = (only_through_mutation * only_not_through_mutation)

                # ------------------------  from dad  ----------------------------

                if parents[1] in one_gene:
                    if parents[0] in one_gene:
                        from_dad = (fifty_fifty * fifty_fifty)
                    elif parents[0] in two_genes:
                        from_dad = (only_through_mutation * fifty_fifty)
                    else:
                        from_dad = (only_not_through_mutation * fifty_fifty)
                elif parents[1] in two_genes:
                    if parents[0] in one_gene:
                        from_dad = (fifty_fifty * only_not_through_mutation)
                    elif parents[0] in two_genes:
                        from_dad = (only_through_mutation * only_not_through_mutation)
                    else:
                        from_dad = (only_not_through_mutation * only_not_through_mutation)
                else:
                    if parents[0] in one_gene:
                        from_dad = (fifty_fifty * only_through_mutation)
                    elif parents[0] in two_genes:
                        from_dad = (only_through_mutation * only_through_mutation)
                    else:
                        from_dad = (only_not_through_mutation * only_through_mutation)

                likelihood_g = (from_mom + from_dad)

            # trait likelihood for one gene
            if person not in have_trait:
                likelihood_t = PROBS['trait'][1][False]
            else:
                likelihood_t = PROBS['trait'][1][True]

        # ------------------------ person has two genes  ----------------------------
        elif person in two_genes:
            # no information on parents
            if parents[0] is None:
                likelihood_g = (PROBS['gene'][2])

            # with information on parents, iterate over options
            else:
                if parents[0] in two_genes:
                    if parents[1] in two_genes:
                        likelihood_g = (only_not_through_mutation * only_not_through_mutation)
                    elif parents[1] in one_gene:
                        likelihood_g = (only_not_through_mutation * fifty_fifty)
                    else:
                        likelihood_g = (only_not_through_mutation * only_through_mutation)
                elif parents[0] in one_gene:
                    if parents[1] in two_genes:
                        likelihood_g = (fifty_fifty * only_not_through_mutation)
                    elif parents[1] in one_gene:
                        likelihood_g = (fifty_fifty * fifty_fifty)
                    else:
                        likelihood_g = (fifty_fifty * only_through_mutation)
                else:
                    if parents[1] in two_genes:
                        likelihood_g = (only_through_mutation * only_not_through_mutation)
                    elif parents[1] in one_gene:
                        likelihood_g = (only_through_mutation * fifty_fifty)
                    else:
                        likelihood_g = (only_through_mutation * only_through_mutation)

            # trait likelihood for 2 genes
            if person not in have_trait:
                likelihood_t = PROBS['trait'][2][False]
            else:
                likelihood_t = PROBS['trait'][2][True]

        joint_probs.append(likelihood_g * likelihood_t)

    p = numpy.prod(joint_probs)
    return p


def update(probabilities, one_gene, two_genes, have_trait, p):
    """
    Add to `probabilities` a new joint probability `p`.
    Each person should have their "gene" and "trait" distributions updated.
    Which value for each distribution is updated depends on whether
    the person is in `have_gene` and `have_trait`, respectively.
    """

    for person in probabilities:

        # Assign the variable gene the correct number depending on what dict they're in
        gene = (1 if person in one_gene
                else 2 if person in two_genes
                else 0)
        
        # Assign the variable trait the correct value depending on the have_trait dict
        trait = (True if person in have_trait else False)
        
        probabilities[person]['gene'][gene] += p
        probabilities[person]['trait'][trait] += p


def normalize(probabilities):
    """
    Update `probabilities` such that each probability distribution
    is normalized (i.e., sums to 1, with relative proportions the same).
    """

    # Iterate over each person
    for person in probabilities:

        g_total = sum(probabilities[person]['gene'].values())
        t_total = sum(probabilities[person]['trait'].values())

        # Change all values according to normalisation with the sum
        probabilities[person]['gene'] = {g: (p/g_total) for g, p in probabilities[person]['gene'].items()}
        probabilities[person]['trait'] = {t: (p/t_total) for t, p in probabilities[person]['trait'].items()}

    return


if __name__ == "__main__":
    main()
